import logging
from typing import Dict, Optional

from ..api import RequestService
from ..endpoints import GenericPluralEndpoint


logger = logging.getLogger(__name__)


class AssetsEndPoint(GenericPluralEndpoint):
    """Endpoint for working with FreshService Assets"""

    def __init__(self, request_service: RequestService, identifier=None):
        super(AssetsEndPoint, self).__init__(request_service=request_service)
        self._endpoint = "/api/v2/assets"
        self.resource_key = "assets"
        self.identifier = identifier
        self._items_per_page = 100

    def delete(
        self, display_id: Optional[int] = None, permanently: Optional[bool] = False
    ) -> Dict:
        """Delete an asset with an option to additionally call the endpoint to permanently delete the item.

        Overriding the inherited method to include the option for a second API request to permanently delete the asset.
        :param display_id: Display ID for the asset to be deleted
        :param permanently: Flag to make a second call to the API to permanently delete the asset
        """
        _method = "DELETE"
        if display_id is not None:
            self.identifier = display_id
        _url = f"{self.item_extended_url}"
        logger.info("Deleting asset with display_id = '%d'", self.identifier)
        response = self.send_request(_url, method=_method)
        if permanently:
            _url = f"{self.item_extended_url}/delete_forever"
            _method = "PUT"
            logger.info(
                "Permanently deleting asset with display_id = '%d'", self.identifier
            )
            response = self.send_request(_url, method=_method)
        return response

    def restore(self, display_id: Optional[int] = None) -> Dict:
        if display_id is not None:
            self.identifier = display_id
        _method = "PUT"
        _url = f"{self.item_extended_url}/restore"
        response = self.send_request(_url, method=_method)
        return response

    def get_associated_requests(self, display_id: Optional[int] = None) -> Dict:
        if display_id is not None:
            self.identifier = display_id
        _url = f"{self.item_extended_url}/requests"
        response = self.send_request(_url)
        return response



from .service_items_endpoint import ServiceItemsEndPoint
from .ticket_form_fields import TicketFormFieldsEndPoint
from .tickets_endpoint import TicketsEndPoint
from .assets import AssetsEndPoint
from .asset_types import AssetTypeEndPoint



import logging

from requests import Session

logger = logging.getLogger(__name__)


class Credential:
    """Facilitate creating a Basic authentication credential for the FreshService API.

    https://api.freshservice.com/#authentication
    """

    def __init__(self, username=None, password=None):
        """Constructor for a Credential object.

        https://api.freshservice.com/#authentication

        :param username: FreshService API key or FreshService Username
        :param password: Dummy password when using an API key. 'X' is used in the documentation
        """
        self.username = username
        self.password = password


class RequestService:
    """Wrapper for the requests package to create an authenticated session with the FreshService API

    Use this class with a context manager or new_session() with a try, except, finally block with
    'RequestService.session.close() in the finally block.
    """

    def __init__(self, credential: Credential, domain: str):
        """Constructor for a RequestService object

        :param credential: Credential object
        :param domain: FreshService domain (the part before '.freshservice.com')
        """
        self.credential = credential
        self.domain = domain
        self.session = None

    def __enter__(self):
        self.new_session()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if any((exc_type, exc_val)):
            logger.exception(
                "Error encountered with requests session. %s - %s",
                exc_type,
                exc_val,
            )
        logger.info("Quitting requests session")
        self.session.close()

    def new_session(self):
        self.session = Session()
        self.session.auth = (self.credential.username, self.credential.password)



from ..endpoints import GenericPluralEndpoint


class TicketsEndPoint(GenericPluralEndpoint):
    def __init__(self, request_service):
        super(TicketsEndPoint, self).__init__(request_service)
        self._endpoint = "/api/v2/tickets"
        self.resource_key = "tickets"



from ..api import RequestService
from ..endpoints import GenericPluralEndpoint


class ServiceItemsEndPoint(GenericPluralEndpoint):
    def __init__(self, request_service: RequestService, display_id=None):
        super(ServiceItemsEndPoint, self).__init__(request_service=request_service)
        self._endpoint = "/api/v2/service_catalog/items"
        self.resource_key = "service_items"
        self.display_id = display_id
        self.create_command = "place_request"

    @property
    def extended_url(self):
        url = super().extended_url
        if self.display_id is not None:
            url = f"{url}/{self.display_id}"
        return url



from ..api import RequestService
from ..endpoints import GenericPluralEndpoint


class TicketFormFieldsEndPoint(GenericPluralEndpoint):
    def __init__(self, request_service: RequestService):
        super(TicketFormFieldsEndPoint, self).__init__(request_service=request_service)
        self._endpoint = "/api/v2/ticket_form_fields"
        self.resource_key = "ticket_fields"



import json
import logging
import os
import sys
from json import JSONDecodeError
from typing import Optional, Dict, Any

from requests import Request
from requests.exceptions import HTTPError

from .api import RequestService

logger = logging.getLogger(__name__)


class GenericEndPoint:
    DEFAULT_HEADERS = {
        "Content-Type": "application/json",
    }

    def __init__(self, request_service: RequestService):
        """Generic class for accessing a FreshService resource.
        """
        self.request_service = request_service
        """instance of RequestService to make the API calls with"""
        self._endpoint = ""
        """string extension from the base of the URL specific to each resource"""
        self.identifier: Any = None
        """Identifier to make the instance of the endpoint specific to a resource."""
        self.resource_key = None
        """Dictionary key used in the return data for a resource.  Used to access the resource in the return data"""
        self.create_command = None
        """Some resources extend the endpoint URL with a verb when creating the resource"""

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def base_url(self):
        return f"https://{self.request_service.domain}.freshservice.com"

    @property
    def extended_url(self):
        return f"{self.base_url}{self.endpoint}"

    @property
    def item_extended_url(self):
        if self.identifier is None:
            return self.extended_url
        else:
            return f"{self.extended_url}/{self.identifier}"

    @property
    def fs_create_requests_enabled(self):
        return (
            True
            if os.getenv("ALLOW_FS_CREATE_REQUESTS", "False").lower() == "true"
            else False
        )

    def get(self, identifier: Any = None) -> Dict:
        """Get a single resource from the FS API

        TODO: Check if identifier is already in the extended_url
        """
        if identifier is not None:
            self.identifier = identifier
        _url = f"{self.item_extended_url}"
        response = self.send_request(_url)
        return response

    def create(self, data: Dict, enabled: Optional[bool] = False) -> Dict:
        url = self.extended_url
        if self.create_command is not None:
            url = f"{url}/{self.create_command}"
        if self.fs_create_requests_enabled or enabled:
            response = self.send_request(url, method="POST", data=data)
        else:
            logger.warning(
                "Environment variable 'ALLOW_FS_CREATE_REQUESTS' must be set to 'True' to allow sending "
                "FreshService create requests or call with create(enabled=True)."
            )
            logger.info(
                "Would have sent 'POST' request to '%s' with data '%s'",
                url,
                json.dumps(data),
            )
            response = {"service_request": {"id": sys.maxsize}}
        return response

    def delete(self, identifier: Any = None) -> Dict:
        """Delete a resource with the FS API

        TODO: Check if the identifier is already in the extended_url
        """
        _method = "DELETE"
        if identifier is not None:
            self.identifier = identifier
        _url = f"{self.extended_url}/{identifier}"
        response = self.send_request(_url, method=_method)
        return response

    def update(self, data: Dict, identifier: Any = None) -> Dict:
        """Update a resource with the FS API

        :param data: Dict with data to update resource.
        :param identifier: Optional identifier to make the endpoint specific to particular resource.
        """
        if identifier is not None:
            self.identifier = identifier
        _method = "PUT"
        _url = f"{self.item_extended_url}"
        response = self.send_request(_url, method=_method, data=data)
        return response

    def send_request(
        self, url: str, method: Optional[str] = "GET", data: Optional[Dict] = None
    ) -> Dict:
        """Send the HTTP request to the FreshService API using a requests library session.

        TODO: Send query strings as a dict for parameters to the requests API.
        """
        try:
            if isinstance(data, dict):
                data = json.dumps(data)
            logger.debug("Generating '%s' request for '%s'", method, url)
            req = Request(method, url, headers=self.DEFAULT_HEADERS, data=data)
            prepped_req = self.request_service.session.prepare_request(req)
            resp = self.request_service.session.send(prepped_req)
            resp.raise_for_status()
            # Not all response objects have json content
            return resp.json()
        except HTTPError as err:
            logger.error("Error encounter with send_request %s", err)
            logger.warning(
                "send_request called to url '%s' with method '%s' and data '%s'",
                url,
                method,
                err.request.body,
            )
            logger.warning(
                "Response: 'status_code' == '%d', 'text' == '%s'",
                err.response.status_code,
                err.response.text,
            )
            raise err
        except JSONDecodeError as excp:
            logger.exception(excp)
            logger.info("Not all response objects have json content.")
            resp_dict = {
                "status_code": getattr(resp, "status_code", None),
                "url": getattr(resp, "url", None),
                "ok": getattr(resp, "ok", None),
                "reason": getattr(resp, "reason", None),
            }
            return resp_dict


class GenericPluralEndpoint(GenericEndPoint):
    DEFAULT_ITEMS_PER_PAGE = 30

    def __init__(self, request_service: RequestService):
        super(GenericPluralEndpoint, self).__init__(request_service)
        self._items_per_page = None

    @property
    def items_per_page(self):
        return (
            self._items_per_page
            if self._items_per_page
            else self.DEFAULT_ITEMS_PER_PAGE
        )

    def get_all(self, query=None):
        """Sends a paginated get request for items of the resource type identified by self.resource_key.
        From the list of dict in the response yields the items selected by self.resource_key.

        Yields a list of dict items from the response selected by self.resource_key until all page results are returned
        in the request.
        TODO: an argument to automatically add "include=type_fields" to the query rather than have the user specifically
            include that.
        """
        page = 1
        url = self.paginate_url(query, page)
        more_results = True
        while more_results:
            result = self.send_request(url)
            items = result.get(self.resource_key)
            if len(items) < self.items_per_page:
                more_results = False
                yield items
            else:
                page += 1
                url = self.paginate_url(query, page)
                yield items

    def paginate_url(self, query=None, page=1):
        """Add page and per_page parameters to the query string.

        TODO: Change this to manipulate a dict and handle pagination with the send_request method and that dict.
        """
        pagination_part = f"page={page}&per_page={self.items_per_page}"
        if query:
            url = f"{self.extended_url}?{pagination_part}&{query}"
        else:
            url = f"{self.extended_url}?{pagination_part}"
        return url



class ServiceRequestModel:
    """
    https://api.freshservice.com/v2/#create_service_request
    {
        "quantity": 1,
        "email": "sample@freshservice.com",
        "custom_fields":{
            "hello" : "test",
            "link" : "https://freshservice.com/"
        },
        "child_items" : [{
            "service_item_id": 22,
            "quantity": 1,
            "custom_fields" : {
                "child_1": "test"
            }
        }]
    }
    """

    def __init__(
        self,
        requester_email: str,
        custom_field_data: dict,
        requested_for=None,
        quantity=1,
    ):
        self.quantity = quantity
        self.email = requester_email
        self.custom_fields = custom_field_data
        self.requested_for = requested_for



"""Top-level package for fshelper."""

__author__ = """Matthew Larsen"""
__email__ = "matt.larsen@connorgp.com"
__version__ = "0.2.1"

from .api import Credential, RequestService
from .v2 import (
    ServiceItemsEndPoint,
    TicketFormFieldsEndPoint,
    TicketsEndPoint,
    AssetsEndPoint,
    AssetTypeEndPoint,
)



#!/usr/bin/env python

from setuptools import setup, find_packages  # type: ignore

with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    author="Matthew Larsen",
    author_email="matt.larsen@connorgp.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    description="FreshService API usage helper",
    install_requires=["requests", ],
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    package_data={},
    include_package_data=True,
    keywords="fshelper",
    name="fshelper",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    setup_requires=[],
    url="https://github.com/matt-larsen-sld/freshservice-helper",
    version="0.2.1",
    zip_safe=False,
)



from typing import Any

from ..api import RequestService
from ..endpoints import GenericPluralEndpoint


class AssetTypeEndPoint(GenericPluralEndpoint):
    def __init__(self, request_service: RequestService, identifier: Any = None):
        super(AssetTypeEndPoint, self).__init__(request_service=request_service)
        self._endpoint = "/api/v2/asset_types"
        self.resource_key = "asset_types"
        self.identifier = identifier
